/*
 Copyright (C) MINZKN.COM
 All rights reserved.
 Code by JaeHyuk Cho <mailto:minzkn@minzkn.com>
*/

/*
   구현시 알아야 하는 주요 용어

   - 선형 회귀
   - 경사 하강법
   - 편미분 / 거듭제곱 / 분산
   - 선형 방정식 (y = w * x + b)

*/

#include <stdio.h>
#include <stdlib.h>

/* 오차 비용이 이 값보다 작은 경우 적중했다고 보고 학습을 종료하는 조건이 됨 */
#define def_hwport_loss_stop 1e-20

/* 입력 데이터 자료형 */
typedef struct { /* H(y) = f(x) */
	double m_x;
	double m_y;
} hwport_data_t;

/* 학습 데이터 자료형 */
typedef struct { /* y = (x * weight) + bias */
	double m_weight; /* for gradient */
	double m_bias;
} hwport_learn_t;

/* 학습 조정 값 자료형 */
typedef struct {
	hwport_learn_t m_initial_learn;
	hwport_learn_t m_step_rate;
	unsigned long m_max_loop;
} hwport_config_t;

static int hwport_dump_data(const hwport_data_t *s_data_ptr, size_t s_count);
static double hwport_pow2(double s_value);
static double hwport_linear_func(double s_weight, double s_x, double s_bias);
static double hwport_gradient_descent(const hwport_config_t *s_config_ptr, hwport_learn_t *s_learn_ptr, const hwport_data_t *s_data_ptr, size_t s_count);
static unsigned long hwport_learning_data(const hwport_config_t *s_config_ptr, hwport_learn_t *s_learn_ptr, const hwport_data_t *s_data_ptr, size_t s_count);
static double hwport_predict_check(const hwport_learn_t *s_learn_ptr, double s_x);

int main(int s_argc, char **s_argv);

/* 입력 데이터를 모두 출력 */
static int hwport_dump_data(const hwport_data_t *s_data_ptr, size_t s_count)
{
	size_t s_index;

	(void)fprintf(stdout, "* Trainging data (count=%lu)\n", s_count);
	for(s_index = (size_t)0ul;s_index < s_count;s_index++) {
		(void)fprintf(
			stdout,
			"\t- data[%lu] : x=%f, y=%f\n",
			s_index,
			s_data_ptr[s_index].m_x,
			s_data_ptr[s_index].m_y
		);
	}
	(void)fprintf(stdout, "\n");

	return(0);
}

/* 거듭 제곱 */
static double hwport_pow2(double s_value)
{
	return(s_value * s_value);
}

/* 선형 함수 H = f(x) = weight * x + bias */
static double hwport_linear_func(double s_weight, double s_x, double s_bias)
{
	return((s_weight * s_x) + s_bias);
}

/* 경사 하강법 */
static double hwport_gradient_descent(const hwport_config_t *s_config_ptr, hwport_learn_t *s_learn_ptr, const hwport_data_t *s_data_ptr, size_t s_count)
{
	size_t s_index;
	double s_hypothesis;
	hwport_learn_t s_delta;
	double s_loss;

	/* 편미분 하여 증감 가중치를 계산 */	
	for(s_index = (size_t)0u, s_delta.m_weight = 0.0, s_delta.m_bias = 0.0;s_index < s_count;s_index++) {
		s_hypothesis = hwport_linear_func(s_learn_ptr->m_weight, s_data_ptr[s_index].m_x, s_learn_ptr->m_bias);
		s_delta.m_weight += (s_hypothesis - s_data_ptr[s_index].m_y) * s_data_ptr[s_index].m_x;
		s_delta.m_bias += s_hypothesis - s_data_ptr[s_index].m_y;
	}
	s_delta.m_weight = (2.0 * s_delta.m_weight) / ((double)s_count);
	s_delta.m_bias = (2.0 * s_delta.m_bias) / ((double)s_count);

	/* 학습된 weight 와 bias 를 편미분 가중치에 따라서 update */
	s_learn_ptr->m_weight -= s_config_ptr->m_step_rate.m_weight * s_delta.m_weight;
	s_learn_ptr->m_bias -= s_config_ptr->m_step_rate.m_bias * s_delta.m_bias;

	/* 입력 데이터와 학습데이터간 오차비용 계산 */
	for(s_index = (size_t)0u, s_loss = 0.0;s_index < s_count;s_index++) {
		s_hypothesis = hwport_linear_func(s_learn_ptr->m_weight, s_data_ptr[s_index].m_x, s_learn_ptr->m_bias);
		s_loss += hwport_pow2(s_hypothesis - s_data_ptr[s_index].m_y); /* pow(s_delta) */
	}

	return(s_loss / ((double)s_count));
}

/* 주어진 입력 데이터 s_data_ptr(data, count)을 주어진 반복학습 횟수 s_count 만큼 선형회귀 학습 */
static unsigned long hwport_learning_data(const hwport_config_t *s_config_ptr, hwport_learn_t *s_learn_ptr, const hwport_data_t *s_data_ptr, size_t s_count)
{
	unsigned long s_loop;
	double s_loss;

	(void)fprintf(stdout, "* Learning...\n");
	(void)fprintf(
		stdout,
	       	"\t- config[initial_learn] : weight=%f, bias=%f\n",
		s_config_ptr->m_initial_learn.m_weight,
		s_config_ptr->m_initial_learn.m_bias
	);
	(void)fprintf(
		stdout,
	       	"\t- config[step_rate] : weight=%f, bias=%f\n",
		s_config_ptr->m_step_rate.m_weight,
		s_config_ptr->m_step_rate.m_bias
	);
	(void)fprintf(
		stdout,
	       	"\t- config[loop] : maximum repeat %lu\n",
		s_config_ptr->m_max_loop
	);

	*s_learn_ptr = s_config_ptr->m_initial_learn;
	for(s_loop = 0ul;s_loop < s_config_ptr->m_max_loop;s_loop++) { /* 반복 학습 loop */
		/* 경사 하강법 계산 */
		s_loss = hwport_gradient_descent(s_config_ptr, s_learn_ptr, s_data_ptr, s_count);

		if((s_loss < def_hwport_loss_stop) /* 적중 */ || ((s_loop + 1lu) == s_config_ptr->m_max_loop) /* 마지막 */ || ((s_loop % (s_config_ptr->m_max_loop / 1000lu)) == 0lu) /* 일정 간격 */) {
			(void)fprintf(stdout,
				"\t\tloop=%lu, loss=%f(%e), weight=%f, bias=%f\n",
			       	s_loop,
			       	s_loss,
			       	s_loss,
			       	s_learn_ptr->m_weight,
			       	s_learn_ptr->m_bias
			);
			if(s_loss < def_hwport_loss_stop) {
				break;
			}
		}
	}

	(void)fprintf(
		stdout,
		"\t- LEARN_RESULT : lean_loop=%lu, learn_weight=%f, learn_bias=%f, loss=%e/%e\n",
		s_loop,
		s_learn_ptr->m_weight,
		s_learn_ptr->m_bias,
		s_loss,
		def_hwport_loss_stop
	);
	
	(void)fprintf(stdout, "\n");

	return(s_loop);
}

/* 주어진 s_x 가 학습된 데이터 s_learn_ptr(weight, bias) 와 계산하여 결과 값 s_predict_y 를 확인 */
static double hwport_predict_check(const hwport_learn_t *s_learn_ptr, double s_x)
{
	double s_predict_y;

	s_predict_y = hwport_linear_func(
		s_learn_ptr->m_weight,
		s_x,
		s_learn_ptr->m_bias
	);

	(void)fprintf(
		stdout,
		"* predict x=%f, y=%f (Y = weight * X + bias = %f * %f + %f)\n",
		s_x,
		s_predict_y,
		s_learn_ptr->m_weight,
		s_x,
		s_learn_ptr->m_bias
	);

	return(s_predict_y);
}

int main(int s_argc, char **s_argv)
{
	/* INPUT : 입력되는 학습할 데이터 (최소한 3개 이상일 때 학습이 될 수 있음) */
	hwport_data_t s_training_data[] = {
		{
			.m_x = 1.0,
			.m_y = (1500.0 * 1.0) + 15.0,
		},
		{
			.m_x = 2.0,
			.m_y = (1500.0 * 2.0) + 15.0,
		},
		{
			.m_x = 3.0,
			.m_y = (1500.0 * 3.0) + 15.0,
		},
		{
			.m_x = 4.0,
			.m_y = (1500.0 * 4.0) + 15.0,
		},
		{
			.m_x = 5.0,
			.m_y = (1500.0 * 5.0) + 15.0,
		},
		{
			.m_x = 6.0,
			.m_y = (1500.0 * 6.0) + 15.0,
		},
		{
			.m_x = 7.0,
			.m_y = (1500.0 * 7.0) + 15.0,
		},
	};
	size_t s_trainging_data_count = (size_t)(sizeof(s_training_data) / sizeof(hwport_data_t));

	/* CONFIG : 학습시 초기 값과 단계 값, 그리고 반복 횟수를 여기서 조정할 수 있습니다. */
	hwport_config_t s_config = {
		.m_initial_learn = { /* 초기 값 */
			.m_weight = 1.0,
			.m_bias = 1.0,
		},
		.m_step_rate = { /* 단계 값 (이 값이 작을 수록 학습 횟수가 많이 필요하지만 정밀도가 향상될 수 있음) */
			.m_weight = 0.01,
			.m_bias = 0.01,
		},
		.m_max_loop = 1000000ul, /* 반복 학습 최대 횟수 */
	};

	/* OUTPUT : learned */
	hwport_learn_t s_learn;

	(void)fprintf(stdout, "* Linear regression test\n\n");

	/* PRINT : 학습할 입력 데이터 확인 */
	(void)hwport_dump_data((const hwport_data_t *)(&s_training_data[0]), s_trainging_data_count);

	/* LEARNING : 학습 */
	(void)hwport_learning_data(
		(const hwport_config_t *)(&s_config),
		(hwport_learn_t *)(&s_learn),
		(const hwport_data_t *)(&s_training_data[0]),
		s_trainging_data_count
	);

	/* VERIFY : 학습한 데이터를 통한 예측이 맞는지 검증 */
	do {
		double s_test_x;

		(void)hwport_dump_data((const hwport_data_t *)(&s_training_data[0]), s_trainging_data_count);
		for(s_test_x = 1.0;s_test_x < 20.0;s_test_x += 1.0) {
			(void)hwport_predict_check((const hwport_learn_t *)(&s_learn), s_test_x);
		}
	}while(0);

	return(EXIT_SUCCESS);
}

/* End of source */
